import math
from collections import OrderedDict
from datetime import date, datetime

import numpy as np
import pandas as pd
import pyarrow as pa
from pandas.core.dtypes.common import is_integer_dtype
from pytest import raises
from triad.collections.schema import Schema, SchemaError
from triad.exceptions import InvalidOperationError, NoneArgumentError


def test_schema_init():
    s = Schema("a:int,b:str")
    assert 2 == len(s)
    assert Schema(s) == "a:int,b:str"
    assert Schema(s) is not s
    assert Schema(a=int, b=str) == "a:long,b:str"
    assert Schema("a:int,b:str") == "a:int,b:str"
    assert Schema("a:int", "b:str") == "a:int,b:str"
    assert Schema(dict(a=int, b="str")) == "a:long,b:str"
    assert (
        Schema(("a", int), [("b", str), ("c", "int")], dict(d=datetime, e="str"))
        == "a:long,b:str,c:int,d:datetime,e:str"
    )
    assert Schema(" a:{x:int32, y:str},b:[datetime]") == "a:{x:int,y:str},b:[datetime]"
    pa_schema = pa.schema([pa.field("123", pa.int32()), pa.field("b", pa.string())])
    raises(SchemaError, lambda: Schema(pa_schema))
    raises(SchemaError, lambda: Schema("a:int", b=str))
    assert 0 == len(Schema())
    assert 0 == len(Schema([]))
    assert 0 == len(Schema(None))
    assert 0 == len(Schema(""))


def test_schema_datetime():
    df = pd.DataFrame(
        [[datetime(2020, 1, 1, 2, 3, 4, 5), date(2020, 2, 2)]], columns=["a", "b"]
    )
    assert Schema(df) == "a:datetime,b:date"


def test_schema_properties():
    s = Schema("a:int,b:str")
    assert ["a", "b"] == s.names
    assert [pa.int32(), pa.string()] == s.types
    assert [pa.field("a", pa.int32()), pa.field("b", pa.string())] == s.fields
    assert (
        pa.schema([pa.field("a", pa.int32()), pa.field("b", pa.string())])
        == s.pyarrow_schema
    )
    assert s.pyarrow_schema == s.pyarrow_schema
    assert pd.api.types.is_integer_dtype(s.pd_dtype["a"])
    assert pd.api.types.is_string_dtype(s.pd_dtype["b"])
    assert s.pandas_dtype == s.pd_dtype


def test_schema_copy():
    a = Schema("a:int,b:str").copy()
    assert isinstance(a, Schema)
    assert a == "a:int,b:str"


def test_schema_setter():
    a = Schema("a:int,b:str")
    with raises(NoneArgumentError):
        a["c"] = None  # None is invalid
    with raises(SchemaError):
        a["b"] = "str"  # Update is not allowed
    with raises(SchemaError):
        a["123"] = "int"  # Col name is invalid
    with raises(SchemaError):
        a["x"] = pa.field("y", pa.int32())  # key!=field.name
    with raises(SchemaError):
        a["y"] = pa.large_binary()  # unsupported types
    a["c"] = str
    a["d"] = pa.field("d", pa.int32())
    assert a == "a:int,b:str,c:str,d:int"


def test_schema_eq():
    s = Schema("a:int,b:str")
    assert s != None
    assert not (s == None)
    assert s == s
    assert s == Schema("a:int,b:str")
    assert not (s == Schema("b:str,a:int"))
    assert s == ["a:int", "b:str"]
    assert s != ["a:long", "b:str"]
    assert not (s == ["a:long", "b:str"])
    assert s == [("a", "int"), ("b", str)]
    assert s == OrderedDict([("a", "int"), ("b", str)])


def test_schema_contains():
    s = Schema("a:int,b:str")
    assert None not in s
    assert s in s
    assert "a" in s
    assert "c" not in s
    assert "a:int" in s
    assert "a:long" not in s
    assert pa.field("a", pa.int32()) in s
    assert pa.field("aa", pa.int32()) not in s
    assert pa.field("a", pa.int64()) not in s
    assert ["a", ("b", str)] in s
    assert ["a", ("b", int)] not in s


def test_schema_append():
    s = Schema()
    s.append(pa.field("a", pa.int32()))
    assert s == "a:int"
    raises(SchemaError, lambda: s.append("b"))
    raises(SchemaError, lambda: s.append(123))
    s.append("b:str")
    assert s == "a:int,b:str"
    s.append(Schema("c:int").pa_schema)
    assert s == "a:int,b:str,c:int"
    s.append("")
    assert s == "a:int,b:str,c:int"
    s.append(" ")
    assert s == "a:int,b:str,c:int"
    df = pd.DataFrame([["a", 1], ["b", 2]], columns=["x", "y"])
    assert Schema(df) == "x:str,y:long"


def test_schema_remove():
    s = Schema("a:int,b:str,c:int")
    t = s.remove(s)
    assert t == ""
    t = s.remove(None)
    assert t == "a:int,b:str,c:int"
    assert t is not s
    t = s.remove("")
    assert t == "a:int,b:str,c:int"
    t = s.remove(" ")
    assert t == "a:int,b:str,c:int"
    t = s.remove("b")
    assert t == "a:int,c:int"
    t = s.remove("b:str")
    assert t == "a:int,c:int"
    t = s.remove("b:str", require_type_match=False)
    assert t == "a:int,c:int"
    t = s.remove("b:int", require_type_match=False)
    assert t == "a:int,c:int"
    t = s.remove("b:str", require_type_match=True)
    assert t == "a:int,c:int"
    raises(SchemaError, lambda: s.remove("b:int", require_type_match=True))
    t = s.remove("b:int", require_type_match=True, ignore_type_mismatch=True)
    assert t == "a:int,b:str,c:int"
    assert t is not s
    t = s.remove(Schema("c:int,b:str"))
    assert t == "a:int"
    t = s.remove(pa.field("c", pa.int32()))
    assert t == "a:int,b:str"
    t = s.remove(["c", "b"])
    assert t == "a:int"
    t = s.remove({"c", "b"})
    assert t == "a:int"
    t = s.remove(["c", ("b", str)])
    assert t == "a:int"
    t = s.remove(["c", [("b", str), "a"]])
    assert t == ""
    raises(SchemaError, lambda: s.remove("x"))
    t = s.remove("x", ignore_key_mismatch=True)
    assert t == s
    assert t is not s


def test_schema_extract():
    s = Schema("a:int,b:str,c:int")
    t = s.extract(s)
    assert t == s
    t = s.extract(None)
    assert t == ""
    t = s.extract("")
    assert t == ""
    t = s.extract(" ")
    assert t == ""
    t = s.extract("b")
    assert t == "b:str"
    t = s.extract("b:str")
    assert t == "b:str"
    t = s.extract("b:str", require_type_match=False)
    assert t == "b:str"
    t = s.extract("b:int", require_type_match=False)
    assert t == "b:str"
    t = s.extract("b:str", require_type_match=True)
    assert t == "b:str"
    raises(SchemaError, lambda: s.extract("b:int", require_type_match=True))
    t = s.extract("b:int", require_type_match=True, ignore_type_mismatch=True)
    assert t == ""
    assert t is not s
    t = s.extract(Schema("c:int,b:str"))
    assert t == "c:int,b:str"
    t = s.extract(pa.field("c", pa.int32()))
    assert t == "c:int"
    t = s.extract(["c", "b"])
    assert t == "c:int,b:str"
    raises(SchemaError, lambda: s.extract({"c", "b"}))
    t = s.extract(["c", ("b", str)])
    assert t == "c:int,b:str"
    t = s.extract(["c", [("b", str), "a"]])
    assert t == "c:int,b:str,a:int"
    raises(SchemaError, lambda: s.extract(["x", "b", "a"]))
    t = s.extract(["x", "b", "a"], ignore_key_mismatch=True)
    assert t == "b:str,a:int"
    t = s.extract("x:int,b:str,a:int", ignore_key_mismatch=True)
    assert t == "b:str,a:int"
    raises(SchemaError, lambda: s.extract("x:int,b:str,a:int"))


def test_schema_update_delete():
    s = Schema("a:int,b:str,c:int")
    with raises(SchemaError):
        del s["a"]
    with raises(SchemaError):
        del s["x"]
    with raises(SchemaError):
        s["a"] = str
    raises(SchemaError, lambda: s.pop("a"))
    raises(SchemaError, lambda: s.popitem("a"))
    raises(SchemaError, lambda: s.update(dict(a=int)))


def test_schema_operators():
    s = Schema("a:int,b:str,c:int")
    s += "d:int"
    t = s + "e:int"
    t += ""
    assert s == "a:int,b:str,c:int,d:int"
    assert t == "a:int,b:str,c:int,d:int,e:int"
    t = s - ""
    assert t == s
    t = s - ["a", "c"]
    assert t == "b:str,d:int"
    with raises(SchemaError):
        t -= "a"
    assert t == "b:str,d:int"


def test_schema_set_ops():
    s = Schema("a:int,b:str,c:int")
    s1 = Schema("a:int,f:str")
    s2 = Schema("a:str,f:str")
    assert s.exclude(s1) == "b:str,c:int"
    raises(SchemaError, lambda: s.exclude(s2))
    assert s.exclude(s2, require_type_match=False) == "b:str,c:int"
    assert s.exclude(s2, ignore_type_mismatch=True) == "a:int,b:str,c:int"

    s3 = Schema("a:int,b:int")
    assert s.intersect(s3) == "a:int"
    assert s.intersect(s3, require_type_match=False) == "a:int,b:str"
    raises(SchemaError, lambda: s.intersect(s3, ignore_type_mismatch=False))
    s4 = Schema("b:int,a1:str,a:int,b1:str")
    assert s4.intersect(s3) == "b:int,a:int"
    assert s4.exclude(s3) == "a1:str,b1:str"
    assert s4.intersect(["b1", "b"], use_other_order=True) == "b1:str,b:int"

    s5 = Schema("e:str,c:int,b:int,d:int")
    assert s.union(s5) == "a:int,b:str,c:int,e:str,d:int"
    raises(SchemaError, lambda: s.union(s5, require_type_match=True))
    assert s.union("e:str") == "a:int,b:str,c:int,e:str"
    assert s == "a:int,b:str,c:int"
    s.union_with("e:str")
    assert s == "a:int,b:str,c:int,e:str"


def test_schema_assert_not_empty():
    raises(SchemaError, lambda: Schema().assert_not_empty())
    raises(SchemaError, lambda: Schema(None).assert_not_empty())
    raises(SchemaError, lambda: Schema([]).assert_not_empty())
    assert Schema("a:int").assert_not_empty() == "a:int"


def test_schema_rename():
    s = Schema("a:int,b:str,c:bool").rename(columns=dict(a="c", c="a"))
    assert s == "c:int,b:str,a:bool"
    s = Schema("a:int,b:str,c:bool").rename(
        columns=dict(a="c", c="a"), ignore_missing=True
    )
    assert s == "c:int,b:str,a:bool"
    raises(SchemaError, lambda: s.rename(dict(x="b")))
    raises(SchemaError, lambda: s.rename(dict(a="b")))
    raises(SchemaError, lambda: s.rename(dict(a=123)))


def test_schema_transform():
    s = Schema("a:int,b:str,c:bool")
    assert s.transform() == Schema()
    assert s.transform(None) == Schema()
    assert s.transform("x:str") == "x:str"
    assert s.transform("*") == s
    assert s.transform("*~x,y") == s
    assert s.transform("*,d:str") == "a:int,b:str,c:bool,d:str"
    assert s.transform("*,d:str - a ") == "b:str,c:bool,d:str"
    assert s.transform("*,d:str - c,,a ") == "b:str,d:str"
    assert s.transform("*,d:str - c-a") == "b:str,d:str"
    assert s.transform("*,d:str ~ c,,a,x ") == "b:str,d:str"
    assert s.transform("*,d:str ~ c-a~x") == "b:str,d:str"
    assert s.transform("* + e:int,b:int,d:str") == "a:int,b:int,c:bool,e:int,d:str"
    # multiple operations will be applied in order
    assert s.transform("*+e:int,b:int-c~x,c") == "a:int,b:int,e:int"
    assert s.transform("* + - ~ ") == s  # no op
    assert s.transform("*", {"d": str}, e=str) == "a:int,b:str,c:bool,d:str,e:str"
    assert s.transform(lambda s: s.fields[0], lambda s: s.fields[2]) == "a:int,c:bool"
    assert s.transform(lambda s: s - ["b"]) == "a:int,c:bool"
    raises(SchemaError, lambda: s.transform("**"))
    raises(SchemaError, lambda: s.transform("*", "*"))
    raises(SchemaError, lambda: s.transform("*-x"))



from typing import Any
from triad.exceptions import NoneArgumentError


def assert_or_throw(bool_exp: bool, exception: Any = None) -> None:
    """Assert on expression and throw custom exception

    :param bool_exp: boolean expression to assert on
    :param exception: a custom Exception instance, or any other object that
        will be stringfied and instantiate an AssertionError, or a function
        that can generate the supported data types

    :Examples:

        .. code-block:: python

            assert_or_throw(True, "assertion error")
            assert_or_throw(False)  # raise AssertionError
            assert_or_throw(False, "assertion error")  # raise AssertionError
            assert_or_throw(False, TypeError("assertion error"))  # raise TypeError

            # Lazy evaluations is useful when constructing the error
            # itself is expensive or error-prone. With lazy evaluations, happy
            # path will be fast and error free.
            def fail():  # a function that is slow and wrong
                sleep(10)
                raise TypeError

            assert_or_throw(True, fail())  # (unexpectedly) raise TypeError
            assert_or_throw(True, fail)  # no exception
            assert_or_throw(True, lambda: "a" + fail())  # no exception
            assert_or_throw(False, lambda: "a" + fail())  # raise TypeError

    """
    if not bool_exp:
        _exception: Any = exception
        if callable(exception):
            _exception = exception()
        if _exception is None:
            raise AssertionError()
        if isinstance(_exception, Exception):
            raise _exception
        if isinstance(_exception, str):
            raise AssertionError(_exception)
        raise AssertionError(str(_exception))


def assert_arg_not_none(obj: Any, arg_name: str = "", msg: str = "") -> None:
    """Assert an argument is not None, otherwise raise exception

    :param obj: argument value
    :param arg_name: argument name, if None or empty, it will use `msg`
    :param msg: only when `arg_name` is None or empty, this value is used

    :raises NoneArgumentError: with `arg_name` or `msg`
    """
    if obj is None:
        if arg_name != "" and arg_name is not None:
            msg = f"{arg_name} can't be None"
        msg = msg or ""
        raise NoneArgumentError(msg)



import math
from collections import OrderedDict
from datetime import date, datetime

import pandas as pd
import pyarrow as pa
import numpy as np
from pytest import raises
from triad.collections.schema import Schema
from triad.constants import FLOAT_INF, FLOAT_NAN, FLOAT_NINF
from triad.exceptions import InvalidOperationError, NoneArgumentError
from triad.utils.pyarrow import _to_pydate, _to_pydatetime, apply_schema
import pickle

"""
None,"1",1,1.1,"2020-01-01","2020-01-01 01:02:03",
pd.NaT,pd.Timestamp,float(nan),float(-inf),float(inf),
true,True,false,False

str,int,double,bool,datetime,date,decimal,dict,list

"""


def test_convert_to_null():
    pdt = pd.Timestamp("2020-01-01T02:03:04")

    _test_convert(None, "null", None)
    _test_convert("1", "null", None)
    _test_convert(pd.NaT, "null", None)
    _test_convert(pdt, "null", None)
    _test_convert(FLOAT_NAN, "null", None)
    _test_convert(FLOAT_INF, "null", None)
    _test_convert(FLOAT_NINF, "null", None)
    _test_convert(True, "null", None)
    _test_convert(False, "null", None)


def test_convert_to_str():
    pdt = pd.Timestamp("2020-01-01T02:03:04")

    _test_convert(None, "str", None)
    _test_convert("1", "str", "1")
    _test_convert(pd.NaT, "str", "NaT")
    _test_convert(pdt, "str", "2020-01-01 02:03:04")
    _test_convert(FLOAT_NAN, "str", "nan")
    _test_convert(FLOAT_INF, "str", "inf")
    _test_convert(FLOAT_NINF, "str", "-inf")
    _test_convert(True, "str", "True")
    _test_convert(False, "str", "False")


def test_convert_to_int():
    pdt = pd.Timestamp("2020-01-01T02:03:04")

    _test_convert(None, "int", None)
    _test_convert("1", "int", 1)
    _test_convert(1, "int", 1)
    _assert_raise("1.1", "int")
    _test_convert(1.1, "int", 1)
    _test_convert(pd.NaT, "int", None)
    _assert_raise(pdt, "int")
    _test_convert(FLOAT_NAN, "int", None)
    _test_convert(np.nan, "int", None)
    _assert_raise(FLOAT_INF, "int")
    _assert_raise(FLOAT_NINF, "int")
    _assert_raise("true", "int")


def test_convert_to_double():
    pdt = pd.Timestamp("2020-01-01T02:03:04")

    _test_convert(None, "double", None)
    _test_convert("1", "double", 1.0)
    _test_convert(1.1, "double", 1.1)
    _test_convert("1.1", "double", 1.1)
    _test_convert(pd.NaT, "double", None)
    _assert_raise(pdt, "double")
    _test_convert(FLOAT_NAN, "double", None)
    _test_convert("nan", "double", None)
    _test_convert("NaN", "double", None)
    _test_convert(FLOAT_INF, "double", FLOAT_INF)
    _test_convert("inf", "double", FLOAT_INF)
    _test_convert("INF", "double", FLOAT_INF)
    _test_convert(FLOAT_NINF, "double", FLOAT_NINF)
    _test_convert("-inf", "double", FLOAT_NINF)
    _test_convert("-INF", "double", FLOAT_NINF)
    _assert_raise("true", "double")


def test_convert_to_decimal():
    _assert_not_supported(None, "decimal(5,2)")


def test_convert_to_bool():
    pdt = pd.Timestamp("2020-01-01T02:03:04")

    _test_convert(None, "bool", None)
    _test_convert(True, "bool", True)
    _test_convert(False, "bool", False)
    _test_convert("true", "bool", True)
    _test_convert("True", "bool", True)
    _test_convert("false", "bool", False)
    _test_convert("False", "bool", False)
    _test_convert(pd.NaT, "bool", None)
    _assert_raise(pdt, "bool")
    _test_convert(FLOAT_NAN, "bool", None)
    _test_convert(np.nan, "bool", None)


def test_convert_to_binary():
    pdt = pd.Timestamp("2020-01-01T02:03:04")

    _test_convert(None, "bytes", None)
    _test_convert(b'\x0e\x15', "bytes", b'\x0e\x15')
    _test_convert(bytearray(b'\x0e\x15'), "bytes", b'\x0e\x15')
    _test_convert(False, "bytes", pickle.dumps(False))
    _test_convert("true", "bytes", pickle.dumps("true"))
    _test_convert(pd.NaT, "bytes", pickle.dumps(pd.NaT))
    _test_convert(pdt, "bytes", pickle.dumps(pdt))
    _test_convert(FLOAT_NAN, "bytes", pickle.dumps(FLOAT_NAN))
    _test_convert(np.nan, "bytes", pickle.dumps(np.nan))


def test_convert_to_datetime():
    pdt = pd.Timestamp("2020-01-01T02:03:04")
    dt = datetime(2020, 1, 1, 2, 3, 4)
    d = date(2020, 1, 1)
    _test_convert(None, "datetime", None)
    _assert_raise("1", "datetime")
    _test_convert("2020-01-01 02:03:04", "datetime", dt)
    _test_convert("2020-01-01", "datetime", datetime(2020, 1, 1))
    _test_convert(pd.NaT, "datetime", None)
    _test_convert(pdt, "datetime", dt)
    assert isinstance(_to_pydatetime(pdt), datetime)
    assert not isinstance(_to_pydatetime(pdt), pd.Timestamp)
    _test_convert(dt, "datetime", dt)
    _test_convert(d, "datetime", datetime(2020, 1, 1))
    _assert_raise(FLOAT_NAN, "datetime")


def test_convert_to_date():
    pdt = pd.Timestamp("2020-01-01T02:03:04")
    dt = datetime(2020, 1, 1, 2, 3, 4)
    d = date(2020, 1, 1)
    _test_convert(None, "date", None)
    _assert_raise("1", "date")
    _test_convert("2020-01-01 02:03:04", "date", d)
    _test_convert("2020-01-01", "date", d)
    _test_convert(pd.NaT, "date", None)
    _test_convert(pdt, "date", d)
    assert isinstance(_to_pydate(pdt), date)
    assert not isinstance(_to_pydate(pdt), pd.Timestamp)
    _test_convert(dt, "date", d)
    _test_convert(d, "date", d)
    _assert_raise(FLOAT_NAN, "date")


def test_convert_to_dict_shallow():
    d = dict(a=1)
    _test_convert(None, "{a:int}", None)
    _test_convert(d, "{a:int}", d)
    _test_convert({}, "{a:int}", {})
    _assert_raise("1", "{a:int}")
    _assert_raise("abc", "{a:int}")
    _assert_raise(pd.NaT, "{a:int}")
    _assert_raise(FLOAT_NAN, "{a:int}")
    _assert_raise(True, "{a:int}")


def test_convert_to_dict_deep():
    d = dict(a=1)
    _test_convert_nested(None, "{a:int}", None)
    _test_convert_nested(d, "{a:int}", d)
    _test_convert_nested({}, "{a:int}", dict(a=None))
    _test_convert_nested(
        '{"b":{"c":["1"]}}', "{a:int,b:{c:[int]}}", dict(a=None, b=dict(c=[1]))
    )
    _assert_raise(dict(a="x"), "{a:int}", True)
    _assert_raise(dict(b=123), "{b:{a:int}}", True)


def test_convert_to_list_shallow():
    d = [1]
    _test_convert(None, "[int]", None)
    _test_convert(d, "[int]", d)
    _test_convert([], "[int]", [])
    _assert_raise("1", "[int]")
    _assert_raise("abc", "[int]")
    _assert_raise(pd.NaT, "[int]")
    _assert_raise(FLOAT_NAN, "[int]")
    _assert_raise(True, "[int]")
    _assert_raise(np.ndarray([1,2]), "[int]", [1,2])


def test_convert_to_list_deep():
    d = ["1"]
    _test_convert_nested(None, "[int]", None)
    _test_convert_nested(d, "[int]", [1])
    _test_convert_nested([], "[int]", [])
    _test_convert_nested(
        '[{"b":{"c":["1"]}}]', "[{a:int,b:{c:[int]}}]", [dict(a=None, b=dict(c=[1]))]
    )
    _assert_raise(["x"], "[int]", True)
    _assert_raise(1, "[int]", True)


def _test_convert(orig, expected_type, expected_value):
    a = [[orig]]
    s = Schema("a:" + expected_type).pa_schema
    x = list(apply_schema(s, a))[0]
    y = list(apply_schema(s, a, copy=False))[0]
    for b in [x, y]:
        if isinstance(expected_value, float) and math.isnan(expected_value):
            assert math.isnan(b[0])
        elif expected_value is pd.NaT:
            assert b[0] is pd.NaT
        else:
            assert expected_value == b[0]
    assert x is not a[0]
    assert y is a[0]


def _test_convert_nested(orig, expected_type, expected_value):
    a = [[orig]]
    s = Schema("a:" + expected_type).pa_schema
    x = list(apply_schema(s, a, deep=True))[0]
    y = list(apply_schema(s, a, copy=False, deep=True))[0]
    for b in [x, y]:
        assert expected_value == b[0]
    assert x is not a[0]
    assert y is a[0]


def _assert_raise(orig, expected_type, deep=False):
    with raises(ValueError):
        a = [[orig]]
        s = Schema("a:" + expected_type).pa_schema
        b = list(apply_schema(s, a, deep=deep))[0]


def _assert_not_supported(orig, expected_type, deep=False):
    with raises(NotImplementedError):
        a = [[orig]]
        s = Schema("a:" + expected_type).pa_schema
        b = list(apply_schema(s, a, deep=deep))[0]



import itertools
from collections import OrderedDict

from pytest import raises
from triad.utils.iter import (
    EmptyAwareIterable,
    Slicer,
    make_empty_aware,
    slice_iterable,
    to_kv_iterable,
)


def test_empty_aware_iterable():
    i = _get_iterable("1,2,3")
    e = make_empty_aware(i)
    assert not e.empty
    assert "1,2,3" == ",".join(e)
    assert e.empty

    i = _get_iterable("1")
    e = EmptyAwareIterable(i)
    assert not e.empty
    assert not e.empty
    assert "1" == ",".join(e)
    assert e.empty

    e = EmptyAwareIterable([])
    assert e.empty
    assert "" == ",".join(e)
    assert e.empty
    raises(StopIteration, lambda: e.peek())

    i = _get_iterable("1,2,3")
    e = EmptyAwareIterable(i)
    assert not e.empty
    assert "1,2" == ",".join(itertools.islice(e, 2))
    assert not e.empty
    assert "3" == ",".join(itertools.islice(e, 2))
    assert e.empty

    i = _get_iterable("1,2,3")
    e = EmptyAwareIterable(iter(i))
    assert not e.empty
    assert "1" == e.peek()
    assert "1,2" == ",".join(itertools.islice(e, 2))
    assert not e.empty
    assert "3" == e.peek()
    assert "3" == ",".join(itertools.islice(e, 2))
    assert e.empty


def test_empty_aware_iterable_recursive():
    i = _get_iterable("1,2,3")
    e = make_empty_aware(i)
    ee = make_empty_aware(_wrap_iterable(e, True))
    assert "1,2,3" == ",".join(ee)
    i = _get_iterable("1,2,3")
    e = make_empty_aware(i)
    ee = make_empty_aware(_wrap_iterable(e, False))
    assert "1t,2t,3t" == ",".join(ee)


def test_to_kv_iterable():
    data1 = [(1, 1), (2, 2)]
    data2 = OrderedDict(data1)
    data3 = [[1, 1], (2, 2)]
    data4 = [[1, 1, 3], (2, 2)]
    data5 = [1, (2, 2)]
    data6 = [(1, 1), (2, 2, 3)]

    assert [] == list(to_kv_iterable(None, none_as_empty=True))
    assert [] == list(to_kv_iterable(None))
    assert [] == list(to_kv_iterable([]))
    raises(ValueError, lambda: list(to_kv_iterable(None, none_as_empty=False)))
    assert data1 == list(to_kv_iterable(data1))
    assert data1 == list(to_kv_iterable(data2))
    assert data1 == list(to_kv_iterable(data3))
    raises(TypeError, lambda: list(to_kv_iterable(data4)))
    raises(TypeError, lambda: list(to_kv_iterable(data5)))
    raises(TypeError, lambda: list(to_kv_iterable(123)))
    raises(ValueError, lambda: list(to_kv_iterable(data6)))


def test_slice_iterable():
    # make sure empty iterable will yield no slice
    ll = list(slice_iterable([], lambda n, c, l: n % 2 == 0))
    assert 0 == len(ll)
    assert_slice("", [], lambda n, c, l: n % 2 == 0, lambda x: x)
    assert_slice("1,2-3,4-5", range(1, 6), lambda n, c, l: n % 2 == 0, lambda x: x)
    assert_slice("1,2,3,4,5", range(1, 6), lambda n, c, l: c < l, lambda x: x)
    assert_slice("1-2-3-4-5", range(1, 6), lambda n, c, l: c > l, lambda x: x)
    # for each slice, only iterate some of them
    assert_slice(
        "1-3-5",
        range(1, 6),
        lambda n, c, l: n % 2 == 0,
        lambda x: itertools.islice(x, 1),
    )
    assert_slice(
        "1,2-3,4-5",
        range(1, 6),
        lambda n, c, l: n % 2 == 0,
        lambda x: itertools.islice(x, 100),
    )
    assert_slice(
        "--", range(1, 6), lambda n, c, l: n % 2 == 0, lambda x: itertools.islice(x, 0)
    )
    n = -1

    def sl(it):
        nonlocal n
        n += 1
        return itertools.islice(it, n)

    assert_slice("-3-5", range(1, 6), lambda n, c, l: n % 2 == 0, sl)


def test_slicer():
    assert_slicer("", [], 1, 0, lambda x: 1)
    assert_slicer("", [], 0, 0, lambda x: 1)
    assert_slicer("", [], 0, 1, lambda x: 1)
    assert_slicer("", [], 1, 1, lambda x: 1)
    assert_slicer(".000", [1, 1, 1], 0, 0, None)
    assert_slicer(".000", [1, 1, 1], None, None, None)
    assert_slicer(".0.1.2", [1, 1, 1], 1, 0, None)
    assert_slicer(".00.1", [1, 1, 1], 2, 0, None)
    assert_slicer(".0.1.2", [1, 1, 1], 0, 1, lambda x: 1)
    assert_slicer(".0.1.2", [1, 1, 1], 0, 1, lambda x: 10)
    assert_slicer(".00.1", [1, 1, 1], 0, 2, lambda x: 1)
    assert_slicer(".0.1.2", [1, 1, 1], 1, 2, lambda x: 1)
    assert_slicer(".00.1", [1, 1, 1], 10, 2, lambda x: 1)
    assert_slicer(".000", [1, 1, 1], 10, 20, lambda x: 1)
    assert_slicer(".0.1.2", [1, 1, 1], 1, "2k", lambda x: 1)
    assert_slicer(".00.1", [1, 1, 1], None, "2k", lambda x: 1024)

    class C(object):
        def __init__(self):
            self.arr = []

        def c(self, no, current, last):
            self.arr.append([current, last])
            return current > last

    c = C()
    assert_slicer("", [], 1, 0, lambda x: 1, c.c)
    assert_slicer("", [], 0, 0, lambda x: 1, c.c)
    assert_slicer("", [], 0, 1, lambda x: 1, c.c)
    assert_slicer("", [], 1, 1, lambda x: 1, c.c)
    assert 0 == len(c.arr)
    assert_slicer(".000", [1, 1, 1], 0, 0, None, c.c)
    c = C()
    assert_slicer(".0.1.2", [1, 2, 3], 0, 0, None, c.c)
    assert [[2, 1], [3, 2]] == c.arr
    c = C()
    assert_slicer(".0.1.2", [1, 0, -1], 1, 0, None, c.c)
    assert [[0, 1], [-1, 0]] == c.arr  # is_boundary must be called anyway
    c = C()
    assert_slicer(".00.1", [1, 0, -1], 2, 0, None, c.c)
    assert [[0, 1], [-1, 0]] == c.arr  # is_boundary must be called anyway
    c = C()
    # size and row counters should reset after slicer taking effect
    assert_slicer(".0.11", [1, 2, 1], 2, 0, None, c.c)
    assert [[2, 1], [1, 2]] == c.arr
    c = C()
    assert_slicer(".00.1", [1, 0, -1], 0, 2, lambda x: 1, c.c)
    assert [[0, 1], [-1, 0]] == c.arr
    c = C()
    assert_slicer(".0.1.2", [1, 1, 1], 1, 2, lambda x: 1, c.c)
    assert [[1, 1], [1, 1]] == c.arr
    c = C()
    # size and row counters should reset after slicer taking effect
    assert_slicer(".0.11", [1, 2, 1], 10, 2, lambda x: 1, c.c)
    assert [[2, 1], [1, 2]] == c.arr
    assert_slicer(".000", [1, 1, 1], 10, 20, lambda x: 1)
    assert_slicer(".0.1.2", [1, 1, 1], 1, "2k", lambda x: 1)
    assert_slicer(".00.1", [1, 1, 1], None, "2k", lambda x: 1024)


def assert_slice(expected, iterable, slicer, slice_proc):
    ll = []
    for x in slice_iterable(iterable, slicer):
        assert not x.empty
        assert isinstance(x, EmptyAwareIterable)
        ll.append(",".join(map(str, slice_proc(x))))
    s = "-".join(ll)
    assert expected == s


def assert_slicer(expected, arr, max_row, max_size, sizer, slicer=None):
    r = []
    n = 0
    c = Slicer(sizer, max_row, max_size, slicer=slicer)
    for chunk in c.slice(arr):
        assert isinstance(chunk, EmptyAwareIterable)
        r.append(".")
        for x in chunk:
            r.append(str(n))
        n += 1
    assert expected == "".join(r)


def _get_iterable(s):
    for ss in s.split(","):
        yield ss


def _make_iterable(n):
    while n > 0:
        yield n
        n -= 1


def _wrap_iterable(it, p):
    if p:
        for x in it:
            yield x
    else:
        for x in it:
            yield str(x) + "t"



import datetime
import importlib
import inspect
from types import ModuleType
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
import six
from ciso8601 import parse_datetime
from triad.utils.assertion import assert_or_throw
from triad.utils.string import assert_triad_var_name

EMPTY_ARGS: List[Any] = []
EMPTY_KWARGS: Dict[str, Any] = {}


def get_caller_global_local_vars(
    global_vars: Optional[Dict[str, Any]] = None,
    local_vars: Optional[Dict[str, Any]] = None,
    start: int = -1,
    end: int = -1,
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Get the caller level global and local variables.

    :param global_vars: overriding global variables, if not None,
      will return this instead of the caller's globals(), defaults to None
    :param local_vars: overriding local variables, if not None,
      will return this instead of the caller's locals(), defaults to None
    :param start: start stack level (from 0 to any negative number),
      defaults to -1 which is one level above where this function is invoked
    :param end: end stack level (from ``start`` to any smaller negative number),
      defaults to -1 which is one level above where this function is invoked
    :return: tuple of `global_vars` and `local_vars`

    :Examples:

        .. code-block:: python

            def caller():
                x=1
                assert 1 == get_value("x")

            def get_value(var_name):
                _, l = get_caller_global_local_vars()
                assert var_name in l
                assert var_name not in locals()
                return l[var_name]

    :Notice:

    This is for internal use, users normally should not call this directly.

    If merging multiple levels, the variables on closer level
    (to where it is invoked) will overwrite the further levels values if there
    is overlap.

    :Examples:

        .. code-block:: python

            def f1():
                x=1

                def f2():
                    x=2

                    def f3():
                        _, l = get_caller_global_local_vars(start=-1,end=-2)
                        assert 2 == l["x"]

                        _, l = get_caller_global_local_vars(start=-2,end=-2)
                        assert 1 == l["x"]

                f2()
            f1()
    """
    assert_or_throw(start <= 0, ValueError(f"{start} > 0"))
    assert_or_throw(end <= start, ValueError(f"{end} > {start}"))
    stack = inspect.currentframe().f_back  # type: ignore
    p = 0
    while p > start and stack is not None:
        stack = stack.f_back
        p -= 1
    g_arr: List[Dict[str, Any]] = []
    l_arr: List[Dict[str, Any]] = []
    while p >= end and stack is not None:
        g_arr.insert(0, stack.f_globals)
        l_arr.insert(0, stack.f_locals)
        stack = stack.f_back
        p -= 1
    if global_vars is None:
        global_vars = {}
        for d in g_arr:
            global_vars.update(d)
    if local_vars is None:
        local_vars = {}
        for d in l_arr:
            local_vars.update(d)
    return global_vars, local_vars  # type: ignore


def str_to_object(
    expr: str,
    global_vars: Optional[Dict[str, Any]] = None,
    local_vars: Optional[Dict[str, Any]] = None,
) -> Any:
    """Convert string expression to object. The string expression must express
    a type with relative or full path, or express a local or global instance without
    brackets or operators.

    :param expr: string expression, see examples below
    :param global_vars: overriding global variables, if None, it will
      use the caller's globals(), defaults to None
    :param local_vars: overriding local variables, if None, it will
      use the caller's locals(), defaults to None
    :return: the

    :raises ValueError: unable to find a matching object

    :Examples:

        .. code-block:: python

            class _Mock(object):
                def __init__(self, x=1):
                    self.x = x

            m = _Mock()
            assert 1 == str_to_object("m.x")
            assert 1 == str_to_object("m2.x", local_vars={"m2": m})
            assert RuntimeError == str_to_object("RuntimeError")
            assert _Mock == str_to_object("_Mock")

    :Notice:

    This function is to dynamically load an object from string expression.
    If you write that string expression as python code at the same location, it
    should generate the same result.
    """
    try:
        for p in expr.split("."):
            assert_triad_var_name(p)
        _globals, _locals = get_caller_global_local_vars(global_vars, local_vars)
        if "." not in expr:
            return eval(expr, _globals, _locals)
        parts = expr.split(".")
        v = _locals.get(parts[0], _globals.get(parts[0], None))
        if v is not None and not isinstance(v, ModuleType):
            return eval(expr, _globals, _locals)
        root = ".".join(parts[:-1])
        return getattr(importlib.import_module(root), parts[-1])
    except ValueError:
        raise  # pragma: no cover
    except Exception:
        raise ValueError(expr)


def str_to_type(
    s: str,
    expected_base_type: Optional[type] = None,
    global_vars: Optional[Dict[str, Any]] = None,
    local_vars: Optional[Dict[str, Any]] = None,
) -> type:
    """Given a string expression, find the first/last type from all import libraries.
    If the expression contains `.`, it's supposed to be a relative or full path of
    the type including modules.

    :param s: type expression, for example `triad.utils.iter.Slicer` or `str`
    :param expected_base_type: base class type that must satisfy, defaults to None
    :param global_vars: overriding global variables, if None, it will
      use the caller's globals(), defaults to None
    :param local_vars: overriding local variables, if None, it will
      use the caller's locals(), defaults to None

    :raises TypeError: unable to find a matching type

    :return: found type
    """
    global_vars, local_vars = get_caller_global_local_vars(global_vars, local_vars)
    try:
        obj = str_to_object(s, global_vars, local_vars)
    except ValueError:
        raise TypeError(f"{s} is not a type")
    assert_or_throw(isinstance(obj, type), TypeError(f"{obj} is not a type"))
    assert_or_throw(
        expected_base_type is None or issubclass(obj, expected_base_type),
        TypeError(f"{obj} is not a subtype of {expected_base_type}"),
    )
    return obj


def str_to_instance(
    s: str,
    expected_base_type: Optional[type] = None,
    args: List[Any] = EMPTY_ARGS,
    kwargs: Dict[str, Any] = EMPTY_KWARGS,
    global_vars: Optional[Dict[str, Any]] = None,
    local_vars: Optional[Dict[str, Any]] = None,
) -> Any:
    """Use :func:'~triad.utils.convert.str_to_type' to find a matching type
    and instantiate

    :param s: see :func:'~triad.utils.convert.str_to_type'
    :param expected_base_type: see :func:'~triad.utils.convert.str_to_type'
    :param args: args to instantiate the type
    :param kwargs: kwargs to instantiate the type
    :param global_vars: overriding global variables, if None, it will
      use the caller's globals(), defaults to None
    :param local_vars: overriding local variables, if None, it will
      use the caller's locals(), defaults to None

    :return: the instantiated the object
    """
    global_vars, local_vars = get_caller_global_local_vars(global_vars, local_vars)
    t = str_to_type(s, expected_base_type, global_vars, local_vars)
    return t(*args, **kwargs)


def to_type(
    s: Any,
    expected_base_type: Optional[type] = None,
    global_vars: Optional[Dict[str, Any]] = None,
    local_vars: Optional[Dict[str, Any]] = None,
) -> type:
    """Convert an object `s` to `type`
    * if `s` is `str`: see :func:'~triad.utils.convert.str_to_type'
    * if `s` is `type`: check `expected_base_type` and return itself
    * else: check `expected_base_type` and return itself

    :param s: see :func:'~triad.utils.convert.str_to_type'
    :param expected_base_type: see :func:'~triad.utils.convert.str_to_type'
    :param global_vars: overriding global variables, if None, it will
      use the caller's globals(), defaults to None
    :param local_vars: overriding local variables, if None, it will
      use the caller's locals(), defaults to None

    :raises TypeError: if no matching type found

    :return: the matching type
    """
    global_vars, local_vars = get_caller_global_local_vars(global_vars, local_vars)
    if isinstance(s, str):
        return str_to_type(s, expected_base_type, global_vars, local_vars)
    if isinstance(s, type):
        if expected_base_type is None or issubclass(s, expected_base_type):
            return s
        raise TypeError(f"Type mismatch {s} expected {expected_base_type}")
    t = type(s)
    if expected_base_type is None or issubclass(t, expected_base_type):
        return t
    raise TypeError(f"Type mismatch {s} expected {expected_base_type}")


def to_instance(
    s: Any,
    expected_base_type: Optional[type] = None,
    args: List[Any] = EMPTY_ARGS,
    kwargs: Dict[str, Any] = EMPTY_KWARGS,
    global_vars: Optional[Dict[str, Any]] = None,
    local_vars: Optional[Dict[str, Any]] = None,
) -> Any:
    """If s is str or type, then use :func:'~triad.utils.convert.to_type' to find
    matching type and instantiate. Otherwise return s if it matches constraints

    :param s: see :func:'~triad.utils.convert.to_type'
    :param expected_base_type: see :func:'~triad.utils.convert.to_type'
    :param args: args to instantiate the type
    :param kwargs: kwargs to instantiate the type
    :param global_vars: overriding global variables, if None, it will
      use the caller's globals(), defaults to None
    :param local_vars: overriding local variables, if None, it will
      use the caller's locals(), defaults to None

    :raises ValueError: if s is an instance but not a (sub)type of `expected_base_type`
    :raises TypeError: if s is an instance, args and kwargs must be empty

    :return: the instantiated object
    """
    global_vars, local_vars = get_caller_global_local_vars(global_vars, local_vars)
    if s is None:
        raise ValueError("None can't be converted to instance")
    if isinstance(s, (str, type)):
        t = to_type(s, expected_base_type, global_vars, local_vars)
        return t(*args, **kwargs)
    else:
        if expected_base_type is not None and not isinstance(s, expected_base_type):
            raise TypeError(f"{str(s)} is not a subclass of {str(expected_base_type)}")
        if len(args) > 0 or len(kwargs) > 0:
            raise ValueError(f"Can't instantiate {str(s)} with different parameters")
        return s


def to_function(
    func: Any,
    global_vars: Optional[Dict[str, Any]] = None,
    local_vars: Optional[Dict[str, Any]] = None,
) -> Any:  # noqa: C901
    """For an expression, it tries to find the matching function.

    :params s: a string expression or a callable
    :param global_vars: overriding global variables, if None, it will
      use the caller's globals(), defaults to None
    :param local_vars: overriding local variables, if None, it will
      use the caller's locals(), defaults to None

    :raises AttributeError: if unable to find such a function

    :return: the matching function
    """
    if isinstance(func, str):
        global_vars, local_vars = get_caller_global_local_vars(global_vars, local_vars)
        try:
            func = str_to_object(func, global_vars, local_vars)
        except ValueError:
            raise AttributeError(f"{func} is not a function")
    assert_or_throw(
        callable(func) and not isinstance(func, six.class_types),
        AttributeError(f"{func} is not a function"),
    )
    return func


def get_full_type_path(obj: Any) -> str:
    """Get the full module path of the type (if `obj` is class or function) or type
    of the instance (if `obj` is an object instance)

    :param obj: a class/function type or an object instance
    :raises TypeError: if `obj` is None, lambda, or neither a class or a function
    :return: full path string
    """
    if obj is not None:
        if inspect.isclass(obj):
            return "{}.{}".format(obj.__module__, obj.__name__)
        if inspect.isfunction(obj):
            if obj.__name__.startswith("<lambda"):
                raise TypeError("Can't get full path for lambda functions")
            return "{}.{}".format(obj.__module__, obj.__name__)
        if isinstance(obj, object):
            return "{}.{}".format(obj.__class__.__module__, obj.__class__.__name__)
    raise TypeError(f"Unable to get type full path from {obj}")


def to_bool(obj: Any) -> bool:
    """Convert an object to python bool value. It can handle values
    like `True`, `true`, `yes`, `1`, etc

    :param obj: object
    :raises TypeError: if failed to convert
    :return: bool value
    """
    if obj is None:
        raise TypeError("None can't convert to bool")
    o = str(obj).lower()
    if o in ["true", "yes", "1"]:
        return True
    if o in ["false", "no", "0"]:
        return False
    raise TypeError(f"{o} can't convert to bool")


def to_datetime(obj: Any) -> datetime.datetime:
    """Convert an object to python datetime. If the object is a
    string, it will use `ciso8601.parse_datetime` to parse

    :param obj: object
    :raises TypeError: if failed to convert
    :return: datetime value
    """
    if obj is None:
        raise TypeError("None can't convert to datetime")
    if isinstance(obj, datetime.datetime):
        return obj
    if isinstance(obj, datetime.date):
        return datetime.datetime(obj.year, obj.month, obj.day)
    if isinstance(obj, str):
        try:
            return parse_datetime(obj)
        except Exception as e:
            raise TypeError(f"{obj} can't convert to datetime", e)
    raise TypeError(f"{type(obj)} {obj} can't convert to datetime")


def to_timedelta(obj: Any) -> datetime.timedelta:
    """Convert an object to python datetime.

    If the object is a string, `min` or `-inf` will return `timedelta.min`,
    `max` or `inf` will return `timedelta.max`; if the object is a number,
    the number will be used as the seconds argument; Otherwise it will use
    `pandas.to_timedelta` to parse the object.

    :param obj: object
    :raises TypeError: if failed to convert
    :return: timedelta value
    """
    if obj is None:
        raise TypeError("None can't convert to timedelta")
    if isinstance(obj, datetime.timedelta):
        return obj
    if np.isreal(obj):
        return datetime.timedelta(seconds=float(obj))
    try:
        return pd.to_timedelta(obj).to_pytimedelta()
    except Exception as e:
        if isinstance(obj, str):
            obj = obj.lower()
            if obj in ["min", "-inf"]:
                return datetime.timedelta.min
            elif obj in ["max", "inf"]:
                return datetime.timedelta.max
        raise TypeError(f"{type(obj)} {obj} can't convert to timedelta", e)


def as_type(obj: Any, target: type) -> Any:
    """Convert `obj` into `target` type

    :param obj: input object
    :param target: target type

    :return: object in the target type
    """
    if issubclass(type(obj), target):
        return obj
    if target == bool:
        return to_bool(obj)
    if target == datetime.datetime:
        return to_datetime(obj)
    if target == datetime.timedelta:
        return to_timedelta(obj)
    return target(obj)


def to_size(exp: Any) -> int:
    """Convert input value or expression to size
    For expression string, it must be in the format of
    `<value>` or `<value><unit>`. Value must be 0 or positive,
    default unit is byte if not provided. Unit can be `b`, `byte`,
    `k`, `kb`, `m`, `mb`, `g`, `gb`, `t`, `tb`.

    Args:
        exp (Any): expression string or numerical value

    Raises:
        ValueError: for invalid expression
        ValueError: for negative values

    Returns:
        int: size in byte
    """
    n, u = _parse_value_and_unit(exp)
    assert n >= 0.0, "Size can't be negative"
    if u in ["", "b", "byte", "bytes"]:
        return int(n)
    if u in ["k", "kb"]:
        return int(n * 1024)
    if u in ["m", "mb"]:
        return int(n * 1024 * 1024)
    if u in ["g", "gb"]:
        return int(n * 1024 * 1024 * 1024)
    if u in ["t", "tb"]:
        return int(n * 1024 * 1024 * 1024 * 1024)
    raise ValueError(f"Invalid size expression {exp}")


def _parse_value_and_unit(exp: Any) -> Tuple[float, str]:
    try:
        assert exp is not None
        if isinstance(exp, (int, float)):
            return float(exp), ""
        exp = str(exp).replace(" ", "").lower()
        i = 1 if exp.startswith("-") else 0
        while i < len(exp):
            if (exp[i] < "0" or exp[i] > "9") and exp[i] != ".":
                break
            i += 1
        return float(exp[:i]), exp[i:]
    except (ValueError, AssertionError):
        raise ValueError(f"Invalid expression {exp}")
